from datetime import date, time
from nmea2000.decoder import NMEA2000Message
from nmea2000.encoder import NMEA2000Encoder
from .test_decoder import _get_decoder

def _validate_129029_message(msg: NMEA2000Message | None):
    assert isinstance(msg, NMEA2000Message)
    assert msg.PGN == 129029
    assert msg.priority == 3
    assert msg.source == 0
    assert msg.destination == 255
    assert msg.description == "GNSS Position Data"
    assert msg.fields[0].value == 231
    assert msg.fields[1].value == date(2013, 3, 1)
    assert msg.fields[2].value == time(19,29,52)
    assert msg.fields[3].value == 42.496768422109845            
    assert msg.fields[4].value == -71.58366365704198
    assert msg.fields[5].value == 90.98460299999999
    assert msg.fields[6].value == 'GPS+SBAS/WAAS'
    assert msg.fields[7].value == 'GNSS fix'
    assert msg.fields[8].value == 'No integrity checking'
    assert msg.fields[9].value == 63
    assert msg.fields[10].value == 8
    assert msg.fields[11].value == 1.11
    assert msg.fields[12].value == 1.9000000000000001
    assert msg.fields[13].value == -33.63
    assert msg.fields[14].value == 0
    assert msg.fields[15].value is None
    assert msg.fields[15].raw_value == 15
    assert msg.fields[16].value is None
    assert msg.fields[17].value is None


def test_decode_strings_from_file_1():
    with open("tests/recombine-frames-1.in", "r") as f:
        lines = f.read().splitlines()

    decoder = _get_decoder()
    counter = 0
    for line in lines:
        input_data = line.strip()
        if input_data.startswith('#') or len(input_data) <= 1:
            continue
        counter += 1
        msg = decoder.decode_basic_string(input_data)
        if counter == 5:
            assert isinstance(msg, NMEA2000Message)
            assert msg.PGN == 130311
            assert msg.priority == 5
            assert msg.source == 35
            assert msg.destination == 255
            assert msg.description == "Environmental Parameters"
            assert msg.fields[0].value == 197
            assert msg.fields[3].value == 281.88
        elif counter == 8:
            _validate_129029_message(msg)
        else:
            assert msg is None
    assert counter == 8

def test_decode_strings_from_file_2():
    with open("tests/recombine-frames-2.in", "r") as f:
        lines = f.read().splitlines()

    decoder = _get_decoder()
    counter = 0
    for line in lines:
        input_data = line.strip()
        if input_data.startswith('#') or len(input_data) <= 1:
            continue
        counter += 1
        msg = decoder.decode_basic_string(input_data)
        if counter == 11:
            _validate_129029_message(msg)
        elif counter == 16:
            assert isinstance(msg, NMEA2000Message)
            assert msg.PGN == 126720
            assert msg.priority == 7
            assert msg.source == 0
            assert msg.destination == 255
            assert msg.description == "0x1EF00: Manufacturer Proprietary fast-packet addressed"
            assert msg.fields[0].value == "Garmin"
            assert msg.fields[1].value == 3
            assert msg.fields[2].value == "Marine"
        else:
            assert msg is None
    assert counter == 16

def test_decode_strings_from_file_2_exclude_id():
    with open("tests/recombine-frames-2.in", "r") as f:
        lines = f.read().splitlines()

    decoder = _get_decoder(exclude_pgns=["0x1ef00ManufacturerProprietaryFastPacketAddressed"])
    counter = 0
    for line in lines:
        input_data = line.strip()
        if input_data.startswith('#') or len(input_data) <= 1:
            continue
        counter += 1
        msg = decoder.decode_basic_string(input_data)
        if counter == 11:
            _validate_129029_message(msg)
        else:
            assert msg is None
    assert counter == 16

def test_decode_strings_from_file_3():
    with open("tests/recombine-frames-3.in", "r") as f:
        lines = f.read().splitlines()

    decoder = _get_decoder()
    counter = 0
    for line in lines:
        input_data = line.strip()
        if input_data.startswith('#') or len(input_data) <= 1:
            continue
        counter += 1
        msg = decoder.decode_basic_string(input_data)
        if counter == 7:
            _validate_129029_message(msg)
        else:
            assert msg is None
    assert counter == 7
    
def test_decode_strings_from_file_4():
    with open("tests/recombine-frames-4.in", "r") as f:
        lines = f.read().splitlines()

    decoder = _get_decoder()
    counter = 0
    for line in lines:
        input_data = line.strip()
        if input_data.startswith('#') or len(input_data) <= 1:
            continue
        counter += 1
        msg = decoder.decode_basic_string(input_data)
        if counter == 13:
            _validate_129029_message(msg)
        else:
            assert msg is None
    assert counter == 13
    
def test_decode_strings_from_file_5():
    with open("tests/recombine-frames-5.in", "r") as f:
        lines = f.read().splitlines()

    decoder = _get_decoder()
    counter = 0
    for line in lines:
        input_data = line.strip()
        if input_data.startswith('#') or len(input_data) <= 1:
            continue
        counter += 1
        msg = decoder.decode_basic_string(input_data)
        if counter == 32:
            assert isinstance(msg, NMEA2000Message)
            assert msg.PGN == 130823
            assert msg.priority == 3
            assert msg.source == 27
            assert msg.destination == 255
            assert msg.description == "Maretron: Proprietary Temperature High Range"
            assert msg.fields[0].id == "manufacturerCode"
            assert msg.fields[0].part_of_primary_key
            assert msg.fields[6].value == 0.1
            assert msg.fields[7].value == 0.4
        else:
            assert msg is None
    assert counter == 32
    
def test_encode_fast():
    json = '{"PGN":129029,"id":"gnssPositionData","description":"GNSS Position Data","fields":[{"id":"sid","name":"SID","description":null,"unit_of_measurement":null,"value":231,"raw_value":231,"physical_quantities":null,"type":[1],"part_of_primary_key":false},{"id":"date","name":"Date","description":null,"unit_of_measurement":"d","value":"2013-03-01","raw_value":15765,"physical_quantities":[18],"type":[9],"part_of_primary_key":false},{"id":"time","name":"Time","description":"Seconds since midnight","unit_of_measurement":"s","value":"19:29:52","raw_value":70192.0,"physical_quantities":[19],"type":[8],"part_of_primary_key":false},{"id":"latitude","name":"Latitude","description":null,"unit_of_measurement":"deg","value":42.496768422109845,"raw_value":42.496768422109845,"physical_quantities":[20],"type":[1],"part_of_primary_key":false},{"id":"longitude","name":"Longitude","description":null,"unit_of_measurement":"deg","value":-71.58366365704198,"raw_value":-71.58366365704198,"physical_quantities":[20],"type":[1],"part_of_primary_key":false},{"id":"altitude","name":"Altitude","description":"Altitude referenced to WGS-84","unit_of_measurement":"m","value":90.98460299999999,"raw_value":90.98460299999999,"physical_quantities":[10],"type":[1],"part_of_primary_key":false},{"id":"gnssType","name":"GNSS type","description":null,"unit_of_measurement":null,"value":"GPS+SBAS/WAAS","raw_value":3,"physical_quantities":null,"type":[4],"part_of_primary_key":false},{"id":"method","name":"Method","description":null,"unit_of_measurement":null,"value":"GNSS fix","raw_value":1,"physical_quantities":null,"type":[4],"part_of_primary_key":false},{"id":"integrity","name":"Integrity","description":null,"unit_of_measurement":null,"value":"No integrity checking","raw_value":0,"physical_quantities":null,"type":[4],"part_of_primary_key":false},{"id":"reserved_258","name":"Reserved","description":null,"unit_of_measurement":null,"value":63,"raw_value":63,"physical_quantities":null,"type":[14],"part_of_primary_key":false},{"id":"numberOfSvs","name":"Number of SVs","description":"Number of satellites used in solution","unit_of_measurement":null,"value":8,"raw_value":8,"physical_quantities":null,"type":[1],"part_of_primary_key":false},{"id":"hdop","name":"HDOP","description":"Horizontal dilution of precision","unit_of_measurement":null,"value":1.11,"raw_value":1.11,"physical_quantities":null,"type":[1],"part_of_primary_key":false},{"id":"pdop","name":"PDOP","description":"Positional dilution of precision","unit_of_measurement":null,"value":1.9000000000000001,"raw_value":1.9000000000000001,"physical_quantities":null,"type":[1],"part_of_primary_key":false},{"id":"geoidalSeparation","name":"Geoidal Separation","description":"Geoidal Separation","unit_of_measurement":"m","value":-33.63,"raw_value":-33.63,"physical_quantities":[10],"type":[1],"part_of_primary_key":false},{"id":"referenceStations","name":"Reference Stations","description":"Number of reference stations","unit_of_measurement":null,"value":0,"raw_value":0,"physical_quantities":null,"type":[1],"part_of_primary_key":false},{"id":"referenceStationType","name":"Reference Station Type","description":null,"unit_of_measurement":null,"value":null,"raw_value":15,"physical_quantities":null,"type":[4],"part_of_primary_key":false},{"id":"referenceStationId","name":"Reference Station ID","description":null,"unit_of_measurement":null,"value":null,"raw_value":null,"physical_quantities":null,"type":[1],"part_of_primary_key":false},{"id":"ageOfDgnssCorrections","name":"Age of DGNSS Corrections","description":null,"unit_of_measurement":"s","value":null,"raw_value":null,"physical_quantities":[19],"type":[8],"part_of_primary_key":false}],"source":0,"destination":255,"priority":3,"timestamp":"2022-09-28T11:36:59.668000"}'
    msg1 = NMEA2000Message.from_json(json)
    encoder = NMEA2000Encoder()
    msg_packets = encoder.encode_ebyte(msg1)
    decoder = _get_decoder()
    for msg_bytes in msg_packets:
        msg2 = decoder.decode_tcp(msg_bytes)
    assert isinstance(msg2, NMEA2000Message)
    _validate_129029_message(msg2)



from __future__ import annotations

import binascii
from datetime import date, datetime, time, timedelta
from dataclasses import dataclass, field
import hashlib
import logging
from typing import Any
import orjson
from .consts import PhysicalQuantities, FieldTypes
from .utils import kelvin_to_celsius, kelvin_to_fahrenheit, mps_to_knots, pascal_to_bar, pascal_to_PSI, radians_to_degrees

logger = logging.getLogger(__name__)

# Helper function
def int_to_bytes(value):
    # Determine the number of bytes needed
    byte_length = (value.bit_length() + 8) // 8 or 1
    return value.to_bytes(byte_length, byteorder="big", signed=False)

# NMEA2000Message class represents a single NMEA 2000 PGN
@dataclass
class NMEA2000Message:
    PGN: int
    id: str = ''
    description: str = ''
    ttl: timedelta | None = None
    fields: list[NMEA2000Field] = field(default_factory=list)
    source: int = 0
    destination: int = 0
    priority: int = 0
    timestamp: datetime = datetime.now()
    source_iso_name: IsoName | None = None
    hash: str | None = None
    
    def add_data(self, src:int, dest: int, priority:int, timestamp: datetime, source_iso_name: IsoName, build_network_map: bool):
        self.source = src
        self.destination = dest
        self.priority = priority
        self.timestamp = timestamp
        self.source_iso_name = source_iso_name
        self.hash = None

        if build_network_map:
            # Using MD5 as we don't need secure hashing and speed matters
            primary_key = f"{self.id}_{self.source_iso_name.name}"
            for nmea_field in self.fields:
                if nmea_field.part_of_primary_key:
                    primary_key += "_" + str(nmea_field.raw_value)
            logger.debug("primary key: %s. iso name: %s", primary_key, self.source_iso_name)
            self.hash = hashlib.md5(primary_key.encode()).hexdigest()


    def apply_preferred_units(self, preferred_units: dict[PhysicalQuantities, str]):
        if len(preferred_units) == 0:
            return
        for f in self.fields:
            if f.physical_quantities == PhysicalQuantities.TEMPERATURE:
                requested_unit = preferred_units.get(PhysicalQuantities.TEMPERATURE, None)
                if requested_unit == "c":
                    f.unit_of_measurement = "C"
                    f.value = kelvin_to_celsius(f.value)
                elif requested_unit == "f":
                    f.unit_of_measurement = "F"
                    f.value = kelvin_to_fahrenheit(f.value)
            if f.physical_quantities == PhysicalQuantities.PRESSURE:
                requested_unit = preferred_units.get(PhysicalQuantities.PRESSURE, None)
                if requested_unit == "bar":
                    f.unit_of_measurement = "Bar"
                    f.value = pascal_to_bar(f.value)
                elif requested_unit == "psi":
                    f.unit_of_measurement = "PSI"
                    f.value = pascal_to_PSI(f.value)
            if f.physical_quantities == PhysicalQuantities.ANGLE:
                requested_unit = preferred_units.get(PhysicalQuantities.ANGLE, None)
                if requested_unit == "deg":
                    f.unit_of_measurement = "Deg"
                    f.value = radians_to_degrees(f.value)
            if f.physical_quantities == PhysicalQuantities.SPEED:
                requested_unit = preferred_units.get(PhysicalQuantities.SPEED, None)
                if requested_unit == "kts":
                    f.unit_of_measurement = "kts"
                    f.value = mps_to_knots(f.value)

    def __repr__(self):
        return f"NMEA2000Message(PGN={self.PGN}, id={self.id}, pri={self.priority}, src={self.source}, source_iso_name={self.source_iso_name}, dest={self.destination}, description={self.description}, fields={self.fields})"

    def to_string_test_style(self):
        fields_str = ', '.join([field.to_string_test_style() for field in self.fields])
        return f"{self.PGN} {self.description}: {fields_str}"

    def to_json(self):
        def default(obj: Any) -> Any:
            type_obj = type(obj)
            if type_obj is bytes:
                return obj.hex()
            raise TypeError
        return orjson.dumps(self.__dict__,   default=default).decode()

    @staticmethod
    def from_json(json_str):
        data = orjson.loads(json_str)
        msg = NMEA2000Message(**data)
        msg.fields = [NMEA2000Field(**field) for field in data.get("fields", [])]
        return msg
    
    def get_field_by_id(self, id: str) -> NMEA2000Field:
        field = next((f for f in self.fields if f.id == id), None)
        if field is None:
            raise ValueError(f"PGN: {self.id}: Field with id '{id}' is missing.")
        return field

    def get_field_int_value_by_id(self, id: str, default_value: int | None = None) -> int:
        field = self.get_field_by_id(id)
        if not isinstance(field.value, int):
            if default_value is not None:
                return default_value
            raise ValueError(f"PGN: {self.id}: Field with id '{id}' is not an integer. It is {type(field.value).__name__}.")
        return field.value

    def get_field_str_value_by_id(self, id: str) -> str:
        field = self.get_field_by_id(id)
        if not isinstance(field.value, str):
            raise ValueError(f"PGN: {self.id}: Field with id '{id}' is not a string. It is {type(field.value).__name__}.")
        return field.value
    
# NMEA2000Field class represents a single NMEA 2000 field
@dataclass
class NMEA2000Field:
    id: str
    name: str | None = None
    description: str | None = None
    unit_of_measurement: str | None = None
    value: str | int | float | bytes | time | date | None = 0
    raw_value: int | float | str | bytes | None = 0
    physical_quantities: PhysicalQuantities | None = None
    type: FieldTypes = FieldTypes.NUMBER
    part_of_primary_key: bool | None = None

    def __repr__(self):
        return f"NMEA2000Field(id={self.id}, name={self.name}, description={self.description}, unit_of_measurement={self.unit_of_measurement}, value={self.value}, raw_value={self.raw_value}, physical_quantities={self.physical_quantities}, type={self.type}, part_of_primary_key = {self.part_of_primary_key})"

    def to_string_test_style(self):
        if isinstance(self.raw_value, int):
            # Convert integer to bytes (big-endian)
            value_bytes = int_to_bytes(self.raw_value)
            # Use binascii.hexlify to convert bytes to hex
            raw_value_hex = binascii.hexlify(value_bytes, " ").decode().upper()
        else:
            raw_value_hex = self.raw_value
        return f'{self.name} = {self.value} (bytes = "{raw_value_hex}")'
    
class LookupFieldTypeEnumeration:
    name: str
    field_type: FieldTypes
    resolution: float | None
    unit: str | None
    bits: int
    lookup_enumeration: str | None

    def __init__(self, name: str, field_type: FieldTypes, resolution: float | None, unit: str | None, bits: int, lookup_enumeration: str | None):
        self.name = name
        self.field_type = field_type
        self.resolution = resolution
        self.unit = unit
        self.bits = bits
        self.lookup_enumeration = lookup_enumeration

@dataclass
class IsoName:
    unique_number: int
    manufacturer_code: str
    device_instance: int
    device_function: str
    device_class: str
    system_instance: int
    industry_group: str
    arbitrary_address_capable: bool
    name: int

    def __init__(self, message: NMEA2000Message, name: int):
        """
        Initialize an IsoName object from an NMEA2000Message.

        Args:
            message: An NMEA2000Message object containing the NAME field data.
            name: The 64-bit NAME field value as an integer.
        """
        self.name = name
        self.unique_number = message.get_field_int_value_by_id('uniqueNumber', 0)
        self.manufacturer_code = message.get_field_str_value_by_id('manufacturerCode')
        self.device_instance = (
            message.get_field_int_value_by_id('deviceInstanceUpper', 0) << 3
        ) | message.get_field_int_value_by_id('deviceInstanceLower', 0)
        self.device_function = message.get_field_str_value_by_id('deviceFunction')
        self.device_class = message.get_field_str_value_by_id('deviceClass')
        self.system_instance = message.get_field_int_value_by_id('systemInstance', 0)
        self.industry_group = message.get_field_str_value_by_id('industryGroup')
        self.arbitrary_address_capable = message.get_field_str_value_by_id('arbitraryAddressCapable') == "Yes"

    def __repr__(self):
        return (
            f"IsoName(unique_number={self.unique_number}, "
            f"manufacturer_code='{self.manufacturer_code}', "
            f"device_instance={self.device_instance}, "
            f"device_function='{self.device_function}', "
            f"device_class='{self.device_class}', "
            f"system_instance={self.system_instance}, "
            f"industry_group='{self.industry_group}', "
            f"arbitrary_address_capable={self.arbitrary_address_capable}, "
            f"name={self.name})"
        )


import asyncio
import logging
import argparse
import time
import math
from typing import List
from nmea2000.ioclient import Type
from nmea2000.message import NMEA2000Message, NMEA2000Field
from nmea2000.encoder import NMEA2000Encoder
from nmea2000.consts import PhysicalQuantities, FieldTypes

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger("NMEA2000TestServer")

class NMEA2000TestServer:
    """Test TCP server that simulates a NMEA2000 gateway."""

    def __init__(self, host, port: int, type: Type):
        """Initialize the test server.

        Args:
            host: The host address to bind to.
            port: The port to listen on.
        """
        self.host = host
        self.port = port
        self.type = type
        self.server = None
        self.clients: List[asyncio.StreamWriter] = []
        self.running = False
        self.encoder = NMEA2000Encoder()

    async def handle_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        """Handle a new client connection."""
        addr = writer.get_extra_info('peername')
        logger.info(f"New connection from {addr}")

        self.clients.append(writer)
        try:
            while self.running:
                try:
                    # Read data from the client (if any)
                    data = await asyncio.wait_for(reader.read(8), timeout=0.1)
                    if data:
                        logger.info(f"Received from {addr}: {data.hex()}")
                        # Echo back the received data
                        writer.write(data)
                        await writer.drain()
                        logger.info(f"Echoed back: {data.hex()}")
                except asyncio.TimeoutError:
                    # Timeout is expected when no data is received
                    pass
                except asyncio.IncompleteReadError:
                    # Client disconnected
                    logger.info(f"Client {addr} disconnected")
                    break
        except Exception as e:
            logger.error(f"Error handling client {addr}: {e}")
        finally:
            # Clean up
            if writer in self.clients:
                self.clients.remove(writer)
            try:
                writer.close()
                await writer.wait_closed()
            except Exception:
                pass
            logger.info(f"Connection from {addr} closed")

    async def send_single_message(self):
        # Generate a test message
        if self.type == Type.EBYTE:
            message = self._generate_test_message()

            # Encode the message using the NMEA2000Encoder
            tcp_data = self.encoder.encode_ebyte(message)[0]
            logger.info(f"Broadcasting message (PGN {message.PGN}): {tcp_data.hex()}")
        elif self.type == Type.ACTISENSE:
            tcp_data = "A000057.055 09FF7 0FF00 3F9FDCFFFFFFFFFF\n".encode('utf-8')
        elif self.type == Type.YACHT_DEVICES:
            tcp_data = "00:01:54.430 R 15F11910 00 00 00 E5 0B 1D FF FF\r\n".encode('utf-8')
        else:
            raise Exception ("Type not supported")
        # Send the encoded message to all connected clients
        await self.send_to_clients(tcp_data)

    async def broadcast_test_messages(self):
        """Broadcast test NMEA2000 messages to all connected clients."""
        while self.running:
            if not self.clients:
                await asyncio.sleep(1)
                continue

            try:
                await self.send_single_message()
            except Exception as e:
                logger.error(f"Error broadcasting message: {e}")

            await asyncio.sleep(1)

    def _generate_test_message(self) -> NMEA2000Message:
        """Generate a test NMEA2000 message."""
        # Example: Generate a vessel heading message (PGN 127250)
        message = NMEA2000Message(
            PGN=127250,
            priority=2,
            source=1,
            destination=255,
            fields=[
                NMEA2000Field(
                    id="sid",
                    name="Sequence ID",
                    description="Sequence ID",
                    unit_of_measurement="",
                    value=0,
                    raw_value=0,
                    physical_quantities=None,
                    type=FieldTypes.NUMBER
                ),
                NMEA2000Field(
                    id="heading",
                    name="Heading",
                    description="Vessel Heading",
                    unit_of_measurement="rad",
                    value=self._generate_heading(),
                    raw_value=self._generate_heading_raw(),
                    physical_quantities=PhysicalQuantities.ANGLE,
                    type=FieldTypes.FLOAT
                ),
                NMEA2000Field(
                    id="deviation",
                    name="Deviation",
                    description="Magnetic Deviation",
                    unit_of_measurement="rad",
                    value=0,
                    raw_value=0,
                    physical_quantities=PhysicalQuantities.ANGLE,
                    type=FieldTypes.FLOAT
                ),
                NMEA2000Field(
                    id="variation",
                    name="Variation",
                    description="Magnetic Variation",
                    unit_of_measurement="rad",
                    value=0,
                    raw_value=0,
                    physical_quantities=PhysicalQuantities.ANGLE,
                    type=FieldTypes.FLOAT
                ),
                NMEA2000Field(
                    id="reference",
                    name="Reference",
                    description="Heading Reference",
                    unit_of_measurement="",
                    value=0,
                    raw_value=0,
                    physical_quantities=None,
                    type=FieldTypes.LOOKUP
                ),
                NMEA2000Field(
                    id="reserved_58",
                    name="Reserved",
                    value=0,
                    raw_value=0,
                    type=FieldTypes.RESERVED
                )
            ]
        )
        return message

    def _generate_heading(self) -> float:
        """Generate a simulated heading value in radians."""
        current_time = time.time()
        heading_degrees = (current_time * 10) % 360  # Simulate heading in degrees
        return math.radians(heading_degrees)

    def _generate_heading_raw(self) -> int:
        """Generate a simulated raw heading value in radians * 10000."""
        return int(self._generate_heading() * 10000)

    async def send_to_clients(self, data: bytes):
        """Send data to all connected clients."""
        disconnected_clients = []
        for writer in self.clients:
            try:
                writer.write(data)
                await writer.drain()
            except Exception as e:
                logger.error(f"Error sending to client: {e}")
                disconnected_clients.append(writer)

        # Remove disconnected clients
        for writer in disconnected_clients:
            if writer in self.clients:
                self.clients.remove(writer)

    async def start(self):
        """Start the test server."""
        self.running = True
        self.server = await asyncio.start_server(
            self.handle_client, self.host, self.port
        )

        addr = self.server.sockets[0].getsockname()
        logger.info(f'NMEA2000 Test Server running on {addr}')

    def start_broadcast(self):
        # Start broadcasting test messages
        asyncio.create_task(self.broadcast_test_messages())

    async def wait(self):
        assert self.server is not None
        async with self.server:
            await self.server.serve_forever()

    async def stop(self):
        """Stop the test server."""
        self.running = False
        if self.server:
            self.server.close()
            await self.server.wait_closed()

        # Close all client connections
        for writer in self.clients:
            try:
                writer.close()
                await writer.wait_closed()
            except Exception:
                pass
        self.clients = []
        logger.info("Server stopped")


async def main():
    """Run the test server."""
    parser = argparse.ArgumentParser(description='NMEA2000 Test TCP Server')
    parser.add_argument('--host', type=str, default='127.0.0.1', help='Host address to bind to')
    parser.add_argument('--port', type=int, default=8881, help='Port to listen on')
    parser.add_argument("--type", type=lambda s: Type[s.upper()], default=Type.ACTISENSE, help="Type of TCP server (e.g. EBYTE or ACTISENSE)")
    args = parser.parse_args()

    server = NMEA2000TestServer(host=args.host, port=args.port, type=args.type)
    await server.start()
    server.start_broadcast()
    try:
        await server.wait()
    except KeyboardInterrupt:
        logger.info("Keyboard interrupt received, shutting down...")
    finally:
        await server.stop()


if __name__ == "__main__":
    asyncio.run(main())


# Standard Library Imports

from datetime import date, timedelta
import struct
import math
from datetime import time
from typing import Optional, Tuple

def kelvin_to_fahrenheit(kelvin):
    """
    Converts temperature from Kelvin to Fahrenheit.
    Returns:
        float: The temperature in Fahrenheit.
    """
    if kelvin is None:
        return None
    
    # Convert Kelvin to Fahrenheit
    fahrenheit = round((kelvin - 273.15) * (9/5) + 32,0)

    return fahrenheit


def kelvin_to_celsius(kelvin):
    """
    Converts temperature from Kelvin to Celsius.
    Returns:
        float: The temperature in Celsius.
    """
    if kelvin is None:
        return None

    # Convert Kelvin to Celsius
    celsius = round(kelvin - 273.15,2)

    return celsius

def pascal_to_bar(pascal):
    if pascal is None:
        return None

    # Convert pascal to bar
    bar = pascal / 100000

    return bar

def pascal_to_PSI(pascal):
    if pascal is None:
        return None

    # Convert pascal to PSI
    PSI = pascal / 6894.76

    return PSI

def mps_to_knots(mps):
    """
    Converts speed from meters per second (m/s) to nautical knots.
    Returns:
        float: The speed in nautical knots, rounded to one decimal place.
    """
    if mps is None:
        return None
    
    # Conversion factor from m/s to knots
    conversion_factor = 3600 / 1852
    
    # Convert m/s to knots and round to one decimal place
    knots = round(mps * conversion_factor, 1)

    return knots


def radians_to_degrees(radians):
    """
    Converts an angle from radians to degrees.
    Returns:
        float: The angle in degrees.
    """
    if radians is None:
        return None

    # Convert radians to degrees
    degrees = round(math.degrees(radians), 0)
    return degrees

def decode_int(data_raw: int, bit_offset: int, bit_length: int):
    data_raw = data_raw >> bit_offset
    # Create a mask with the desired number of bits set to 1
    mask = (1 << bit_length) - 1
    # Perform bitwise AND with the mask
    result = data_raw & mask
    return result

def decode_date(days_since_epoch: int | float | None) -> date | None:
    """
    Decodes an integer representing the number of days since 1970-01-01 (UNIX epoch)
    """
    if days_since_epoch is None:
        return None
    
    # Ensure the input is treated as an integer
    days_since_epoch = int(days_since_epoch)
    
    # Define the start date as 1970-01-01
    start_date = date(1970, 1, 1)
    
    # Calculate the decoded date by adding the days to the start date
    decoded_date = start_date + timedelta(days=days_since_epoch)

    # Format and return the date string
    return decoded_date

def encode_date(decoded_date: date) -> int:
    """
    Encodes a date into an integer representing the number of days since 1970-01-01 (UNIX epoch)
    """
    # Define the start date as 1970-01-01
    start_date = date(1970, 1, 1)
    
    # Calculate the number of days since the start date
    days_since_epoch = (decoded_date - start_date).days

    return days_since_epoch


def decode_time(seconds_since_midnight: int | float | None) -> time | None:
    """
    Decodes an integer representing seconds since midnight into a time object.
    Returns:
        time: The decoded time.
    """
    if seconds_since_midnight is None:
        return None

    # Ensure the input is treated as an integer
    seconds_since_midnight = int(seconds_since_midnight)

    # Validate input
    if not (0 <= seconds_since_midnight < 86400):  # There are 86400 seconds in a day
        return time(hour=0, minute=0, second=0)

    hours = seconds_since_midnight // 3600  # 3600 seconds in an hour
    minutes = (seconds_since_midnight % 3600) // 60
    seconds = seconds_since_midnight % 60

    return time(hour=hours, minute=minutes, second=seconds)

def encode_time(time: time, bit_length: int) -> int:
    """
    Encodes a time object into an integer representing the number of seconds since midnight.
    Returns:
        int: The number of seconds since midnight.
    """
    if time is None:
        # Set to "not available" value
        return (1 << bit_length) - 1

    # Calculate the number of seconds since midnight
    seconds_since_midnight = time.hour * 3600 + time.minute * 60 + time.second

    return seconds_since_midnight



def decode_decimal(number_int):
    """
    Decodes a numeric value where each byte represents 2 decimal digits in BCD format.
    """
    if number_int is None:
        return None

    decimal_value = 0
    multiplier = 1

    while number_int > 0:
        byte = number_int & 0xFF
        decimal_value += ((byte >> 4) * 10 + (byte & 0x0F)) * multiplier
        number_int >>= 8
        multiplier *= 100

    return decimal_value

def encode_decimal(decimal_value):
    """
    Encodes a numeric value into BCD format where each byte represents 2 decimal digits.
    """
    if decimal_value is None:
        return None

    number_int = 0
    shift = 0

    while decimal_value > 0:
        two_digits = decimal_value % 100
        bcd_byte = ((two_digits // 10) << 4) | (two_digits % 10)
        number_int |= (bcd_byte << shift)
        decimal_value //= 100
        shift += 8

    return number_int

def decode_float(data_raw: int, bit_offset: int, bit_length: int):
    """
    Decodes a 32-bit integer representing an IEEE-754 floating-point number in little endian format into a Python float.
    """
    number_int = decode_int(data_raw, bit_offset, bit_length)
    # Ensure the input integer fits in 32 bits
    if not (0 <= number_int <= 0xFFFFFFFF):
        return 0

    # Convert the integer to bytes. The '<I' format specifies little-endian unsigned 32-bit integer.
    bytes_data = struct.pack('<I', number_int)

    # Unpack the bytes to a float using the '<f' format which specifies little-endian 32-bit floating point.
    decoded_float, = struct.unpack('<f', bytes_data)
    
    return decoded_float

def encode_float(float_number) -> int:
    """
    Encodes a Python float into a 32-bit integer representing an IEEE-754 floating-point number in little endian format.
    """
    if float_number is None:
        raise ValueError("Cannot encode None as a float")

    # Pack the float into bytes using the '<f' format which specifies little-endian 32-bit floating point.
    bytes_data = struct.pack('<f', float_number)

    # Unpack the bytes to an integer using the '<I' format which specifies little-endian unsigned 32-bit integer.
    encoded_int, = struct.unpack('<I', bytes_data)

    return encoded_int


def decode_number(data_raw: int, bit_offset: int, bit_length: int, signed: bool, resolution: float) -> Optional[float]:
    """
    The function follows specific decoding rules based on the bit length of the number:
    - For numbers using 2 or 3 bits, the maximum value indicates the field is not present (None is returned).
    - For numbers using 4 bits or more, the maximum positive value indicates the field is not present (None is returned).
    """
    number_int = decode_int(data_raw, bit_offset, bit_length)

    #make it signed using sign extension operation
    if signed:
        signed_mask = 1 << (bit_length -1)
        if number_int & signed_mask != 0:
            number_int -= (1 << bit_length)

    if bit_length <= 3:
        if number_int == (1 << bit_length) - 1:
            return None
    elif bit_length >= 4:
        max_positive_value = (1 << bit_length) - 1 if not signed else (1 << (bit_length - 1)) - 1
        if number_int == max_positive_value:
            return None

    # adjust resolution
    number_int *= resolution

    return number_int

def encode_number(
    value: Optional[float],
    bit_length: int,
    signed: bool,
    resolution: float
) -> int:
    """
    Encodes a number into a bitfield within an integer.

    - If value is None, the field is set to the "not available" value per bit length.
    - Applies resolution scaling and sign encoding.
    - Modifies the bits in `data_raw` at the specified offset and returns the new value.
    """
    if value is None:
        # Set to "not available" value
        if bit_length <= 3:
            return (1 << bit_length) - 1
        elif signed:
            return (1 << (bit_length - 1)) - 1
        else:
            return (1 << bit_length) - 1

    # Scale using resolution
    number_int = int(round(value / resolution))

    # Check bounds
    if signed:
        min_val = -(1 << (bit_length - 1))
        max_val = (1 << (bit_length - 1)) - 2  # reserve max for "not available"
    else:
        min_val = 0
        max_val = (1 << bit_length) - 2  # reserve max for "not available"

    if not (min_val <= number_int <= max_val):
        raise ValueError(f"Value {value} out of range after scaling")

    # Handle sign bit if negative
    if signed and number_int < 0:
        number_int = (1 << bit_length) + number_int

    return number_int

def decode_bit_lookup(data_raw: int, bit_lookup_dict) -> str:
    bit = 0
    flags = []
    while data_raw !=0:
        if data_raw & 1 == 1:
            str = bit_lookup_dict.get(bit, None)
            if str is not None:
                flags.append(str)
        bit +=1
        data_raw >>= 1
    return ', '.join(flags)

def decode_string_fix(data_raw: int, bit_offset: int, bit_length: int) -> str:
    number_int = decode_int(data_raw, bit_offset, bit_length)
    num_bytes = (bit_length + 7) // 8
    byte_arr = number_int.to_bytes(num_bytes, 'little')
    decoded_str = byte_arr.decode('utf-8', errors='ignore')
    decoded_str = decoded_str.split('\x00', 1)[0]
    decoded_str = decoded_str.split('\xff', 1)[0]
    decoded_str = decoded_str.split('@', 1)[0]
    decoded_str = decoded_str.strip()
    return decoded_str
    
def decode_string_lz(data_raw: int, bit_offset: int) -> str:
    data_raw = data_raw >> bit_offset
    byte_arr = data_raw.to_bytes((data_raw.bit_length() + 7) // 8, byteorder="little")
    str_len = byte_arr[0]
    byte_arr_str = byte_arr[1 : 1 + str_len]
    decoded_str = byte_arr_str.decode('utf-8', errors='ignore')
    return decoded_str
    
def decode_string_lau(data_raw: int, bit_offset: int) -> Tuple[str | None, int]:
    data_raw = data_raw >> bit_offset
    byte_arr = data_raw.to_bytes(((data_raw.bit_length() + 7) // 8)+1, byteorder='little')
    if len(byte_arr) < 2:
        return None, len(byte_arr)
    str_len = byte_arr[0]
    is_asci = byte_arr[1]
    byte_arr_str = byte_arr[2 : str_len]
    if is_asci:
        decoded_str = byte_arr_str.decode('utf-8', errors='ignore')
    else:
        decoded_str = byte_arr_str.decode('utf-16', errors='ignore')
    return decoded_str, str_len*8
    



import logging
import binascii
import os
from datetime import datetime, timedelta
from typing import Tuple
from .message import IsoName, NMEA2000Message
from .pgns import *  # noqa: F403
from .consts import PhysicalQuantities

logger = logging.getLogger(__name__)

class fast_pgn_metadata():
    def __init__(self) -> None:
        self.frames = {}
        self.payload_length = 0
        self.bytes_stored = 0
        self.sequence_counter = -1

    def __repr__(self):
        return f"<fast_pgn_metadata frames={len(self.frames)} payload_length={self.payload_length} bytes_stored={self.bytes_stored} sequence_counter={self.sequence_counter}>"

ISO_CLAIM_PGN = 60928
ISO_CLAIM_PGN_ID = "isoAddressClaim"

class NMEA2000Decoder():
    def __init__(self,
                 exclude_pgns:list[int | str]=[],
                 include_pgns:list[int | str]=[],
                 exclude_manufacturer_code:list[str]=[],
                 include_manufacturer_code:list[str]=[],
                 preferred_units:dict[PhysicalQuantities, str]={},
                 dump_to_file: str | None = None,
                 dump_pgns:list[int | str]=[],
                 build_network_map: bool = False,
                 ) -> None:
        self.data = {}
        self.dump_TextIOWrapper = None
        self.build_network_map = build_network_map
        self.started_at = datetime.now()
        if dump_to_file:
            os.makedirs(os.path.dirname(dump_to_file), exist_ok=True)
            self.dump_TextIOWrapper = open(dump_to_file, 'a')
        if not isinstance(exclude_pgns, list):
            raise ValueError("exclude_pgns must be a list")
        if not isinstance(include_pgns, list):
            raise ValueError("include_pgns must be a list")
        if len(exclude_pgns) > 0 and len(include_pgns) > 0:
            raise ValueError("Only one of exclude_pgns or include_pgns can be used")

        self.exclude_pgns, self.exclude_pgns_ids = self.split_pgn_list(exclude_pgns)
        self.include_pgns, self.include_pgns_ids = self.split_pgn_list(include_pgns)
        self.exclude_manufacturer_code = {k.lower() for k in exclude_manufacturer_code}
        self.include_manufacturer_code = {k.lower() for k in include_manufacturer_code}
        self.dump_include_pgns, self.dump_include_pgns_ids = self.split_pgn_list(dump_pgns)
        self.preferred_units = {k: v.lower() for k, v in preferred_units.items()}
        self.source_to_iso_name: dict[int, IsoName] = {}

        self.iso_claim_filter = (ISO_CLAIM_PGN in self.exclude_pgns) or ("isoAddressClaim" in self.exclude_pgns_ids)
        if self.iso_claim_filter:
            while ISO_CLAIM_PGN in self.exclude_pgns:
                self.exclude_pgns.remove(ISO_CLAIM_PGN)    
            while ISO_CLAIM_PGN_ID in self.exclude_pgns_ids:
                self.exclude_pgns_ids.remove(ISO_CLAIM_PGN_ID)    
            logger.info("iso address claim will be removed later")

        logger.info("PGN filter exclude: %s, %s", self.exclude_pgns, self.exclude_pgns_ids)
        logger.info("PGN filter include: %s, %s", self.include_pgns, self.include_pgns_ids)
        logger.info("Preffered units: %s", self.preferred_units)
        logger.info("Dump location: %s, PGNs: %s", dump_to_file, dump_pgns)

    @staticmethod
    def split_pgn_list(pgn_list: list[int | str]) -> Tuple[list[int], list[str]]:
        """Split a list of PGNs into two lists: one for integers and one for strings."""
        int_list = []
        str_list = []
        for pgn in pgn_list:
            if isinstance(pgn, int):
                int_list.append(pgn)
            elif isinstance(pgn, str):
                str_list.append(pgn.lower())
            else:
                raise ValueError(f"Invalid PGN type: {type(pgn)}. Must be int or str.")
        return int_list, str_list

    def _decode_fast_message(self, pgn, priority, src, dest, timestamp, can_data, source_iso_name: IsoName | None) -> NMEA2000Message | None:
        """Parse a fast packet message and store the data until all frames are received."""
        fast_packet_key = f"{pgn}_{src}_{dest}"
        
        # Check if this PGN already has a storage structure; if not, create one
        if self.data.get(fast_packet_key) is None:
            self.data[fast_packet_key] = fast_pgn_metadata()
            
        fast_pgn = self.data[fast_packet_key]

        # the last byte has the sequence_counter and frame_counter
        last_byte = can_data[-1]
        
        # Extract the sequence counter (high 3 bits) and frame counter (low 5 bits) from the last byte
        sequence_counter = (last_byte >> 5) & 0b111  # Extract high 3 bits
        frame_counter = last_byte & 0b11111  # Extract low 5 bits
        
        total_bytes = None
        
        if frame_counter != 0 and fast_pgn.payload_length == 0:
            logger.debug(f"Ignoring frame {frame_counter} for PGN {pgn} as first frame has not been received.")
            return None
        
        # if this is the first frame of new sequence we will start over
        if frame_counter == 0 and sequence_counter != fast_pgn.sequence_counter:
            
            # Extract the total number of frames from the second-to-last byte
            total_bytes = can_data[-2]
           
            # Start a new pgn hass structure 
        
            fast_pgn.payload_length = total_bytes
            fast_pgn.sequence_counter = sequence_counter
            fast_pgn.bytes_stored = 0  # Reset bytes stored for a new message
            fast_pgn.frames.clear()  # Clear previous frames
                    
            # For the first frame, exclude the last 4 hex characters (2 bytes) from the payload
            data_payload = can_data[:-2]
            
        else:       
            if sequence_counter != fast_pgn.sequence_counter:
                logger.debug(f"Ignoring frame {sequence_counter} for PGN {pgn} as it does not match current sequence.")
                return None
            elif frame_counter in fast_pgn.frames:
                logger.debug(f"Frame {frame_counter} for PGN {pgn} is already stored.")
                return None
            else:
                # For subsequent frames, exclude the last byte from the payload
                data_payload = can_data[:-1]
        
        byte_length = len(data_payload)

        # Store the frame data
        fast_pgn.frames[frame_counter] = data_payload
        fast_pgn.bytes_stored += byte_length  # Update the count of bytes stored
        
        # Log the extracted values
        logger.debug(f"Sequence Counter: {sequence_counter}, Frame Counter: {frame_counter}")
        
        if total_bytes is not None:
            logger.debug(f"Total Payload Bytes: {total_bytes}")

        logger.debug(f"Orig Payload (hex): {can_data}, Data Payload (hex): {data_payload}")
        
        logger.debug("PGN Data: %s", fast_pgn)
        
        # Check if all expected bytes have been stored
        if fast_pgn.bytes_stored >= fast_pgn.payload_length:
            
            logger.debug("All Fast packet frames collected for PGN: %d", pgn)

            # All data for this PGN has been received, proceed to publish
            combined_payload = bytes([b for idx in sorted(fast_pgn.frames) for b in fast_pgn.frames[idx][::-1]])[::-1]
            
            nmea = None
            if combined_payload is not None:
                logger.debug(f"Combined Payload (hex): {combined_payload})")
                nmea = self._call_decode_function(pgn, priority, src, dest, timestamp, combined_payload, source_iso_name)

            # Reset the structure for this PGN
            del self.data[fast_packet_key]
            return nmea
        else:
            logger.debug(f"Waiting for {fast_pgn.payload_length - fast_pgn.bytes_stored} more bytes.")
            return None

    def decode_actisense_string(self, actisense_string: str) -> NMEA2000Message | None:
        """Process an Actisense packet string and extract the PGN, source ID, and CAN data. Based on: https://actisense.com/knowledge-base/nmea-2000/w2k-1-nmea-2000-to-wifi-gateway/nmea-2000-ascii-output-format/"""
        # Split the Actisense string by spaces
        parts = actisense_string.split()
        
        if len(parts) < 3:
            raise ValueError("Invalid Actisense string format")
        
        if not parts[0].startswith("A"):
            raise ValueError("Invalid format: should start with 'A'")
    
        # Extract the timestamp from the first part
        seconds, milliseconds = map(int, parts[0][1:].split("."))
        offset = timedelta(seconds=seconds, milliseconds=milliseconds)
        timestamp = datetime.now() + offset

        # Extract the priority, destination, and source from the second part
        n = int(parts[1], 16)
        priority = n & 0xF
        dest = (n >> 4) & 0xFF
        src = (n >> 12) & 0xFF
        
        # Extract the PGN from the third part
        pgn = int(parts[2], 16)
        
        # Extract the CAN data from the remaining parts
        # Convert to bytes
        bytes_data = bytes.fromhex(parts[3])

        # Reverse the byte order
        reversed_bytes = bytes_data[::-1]

        # Log the extracted information
        logger.debug(f"Priority: {priority}, Destination: {dest}, Source: {src}, PGN: {pgn}, CAN Data: {reversed_bytes}")
        
        return self._decode(pgn, priority, src, dest, timestamp, bytes(reversed_bytes), True)

    def decode_yacht_devices_string(self, yd_string: str) -> NMEA2000Message | None:
        """Process an Yacht Devices string and extract the PGN, source ID, and CAN data. Based on: https://www.yachtd.com/downloads/ydwg02.pdf page 62-63"""
        # Split the Actisense string by spaces
        parts = yd_string.split()
        
        if len(parts) < 4:
            raise ValueError("Invalid Yacht Devices string format")
        
        if parts[1] not in ["R", "T"]:
            raise ValueError("Invalid format: 2nd part should be 'R'")
    
        # Extract the timestamp from the first part
        timestamp = datetime.strptime(parts[0], "%H:%M:%S.%f")

        # Extract the PGN, priority, destination, and source from the second part
        msgid = int(parts[2], 16)
        pgn_id, source_id, dest, priority = NMEA2000Decoder._extract_header(msgid)
        
        # Extract the CAN data from the remaining parts
        can_data = parts[3:][::-1]
        can_data_bytes = [int(byte, 16) for byte in can_data]

        # Log the extracted information
        logger.debug(f"Priority: {priority}, Destination: {dest}, Source: {source_id}, PGN: {pgn_id}, CAN Data: {can_data_bytes}")
        
        return self._decode(pgn_id, priority, source_id, dest, timestamp, bytes(can_data_bytes))

    def decode_basic_string(self, basic_string: str, already_combined: bool = False) -> NMEA2000Message | None:
        """Process an Actisense packet string and extract the PGN, source ID, and CAN data."""
        # Split the Actisense string by spaces
        parts = basic_string.split(",")
        
        if len(parts) < 7: # should have at least one data bytes probably
            raise ValueError("Invalid string format")
        
        # Extract the fields
        if parts[0].endswith("Z"):
            timestamp = datetime.strptime(parts[0], "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            timestamp = datetime.strptime(parts[0], "%Y-%m-%d-%H:%M:%S.%f")
        priority = int(parts[1])
        pgn_id = int(parts[2])
        src = int(parts[3])
        dest = int(parts[4])
        length = int(parts[5])
        # Extract the CAN data from the remaining parts
        can_data = parts[6:6 + length][::-1]
        can_data_bytes = [int(byte, 16) for byte in can_data]
        # Log the extracted information
        logger.debug(f"Priority: {priority}, Destination: {dest}, Source: {src}, PGN: {pgn_id}, CAN Data: {can_data_bytes}")
        
        # not calling _decode as in this format the fast frames are already combined
        return self._decode(pgn_id, priority, src, dest, timestamp, bytes(can_data_bytes), already_combined)

    @staticmethod
    def _extract_header(frame_id_int: int) -> Tuple[int, int, int, int]:
        """
        Extracts PGN, source ID, destination, and priority from a 29-bit CAN frame ID.
        Returns a tuple of (pgn_id, source_id, dest, priority).
        based on the 29 bits (ID0 - ID28) in https://canboat.github.io/canboat/canboat.html
        """
        source_id = frame_id_int & 0xFF             # bits 0-7 = 8 bits
        pgn_id_raw = (frame_id_int >> 8) & 0x3FFFF  # bits 8-25 = 18 bits
        priority = (frame_id_int >> 26) & 0x07      # bits 26-28 = 3 bits

        dp = (pgn_id_raw >> 16) & 0x3    # bits 16-17
        pf = (pgn_id_raw >> 8) & 0xFF    # bits 8-15
        ps = pgn_id_raw & 0xFF           # bits 0-7

        if pf < 0xF0:
            # PDU1 format: PS is destination address
            dest = ps
            pgn_id = (dp << 16) | (pf << 8)
        else:
            # PDU2 format: broadcast, destination is always 255
            dest = 255
            pgn_id = (dp << 16) | (pf << 8) | ps

        return pgn_id, source_id, dest, priority

    def decode_tcp(self, packet: bytes) -> NMEA2000Message | None:
        """Tested with ECAN devices. Process a single packet and extract the PGN, source ID, and CAN data."""
        
        # First byte has the data length in the lowest 4 bits
        type_byte = packet[0]
        data_length = type_byte & 0x0F  # last 4 bits represent the data length
        
        # Extract the frame ID
        frame_id = packet[1:5]      
        # Convert frame_id bytes to an integer
        frame_id_int = int.from_bytes(frame_id, byteorder='big')        
        # Parse it
        pgn_id, source_id, dest, priority = NMEA2000Decoder._extract_header(frame_id_int)

        # Extract and reverse the CAN data
        can_data = packet[5:5 + data_length][::-1]
               
        # Log the extracted information including the combined string
        logger.debug("PGN ID: %s, Frame ID: %s, CAN Data: %s, Source ID: %s",
            pgn_id,
            binascii.hexlify(frame_id).decode('ascii'),
            can_data,
            source_id)
        
        return self._decode(pgn_id, priority, source_id, dest, datetime.now(), bytes(can_data))

    def decode_usb(self, packet: bytes) -> NMEA2000Message | None:
        """Tested with Waveshare-usb-a device. Process a single packet and extract the PGN, source ID, and CAN data."""
        if packet[0] != 0xaa or packet[-1 ] != 0x55:
            raise Exception ("Packet does not have the right prefix and suffix")
        
        if len(packet) < 2 + 4 + 1: # 2 headers, 4 id, 1 data
            raise Exception ("Packet is too short")    
        
        # First byte has the data length in the lowest 4 bits
        type_byte = packet[1]
        data_length = type_byte & 0x0F  # last 4 bits represent the data length
        
        # Extract the frame ID
        frame_id = packet[2:6]        
        # Convert frame_id bytes to an integer
        frame_id_int = int.from_bytes(frame_id, byteorder='little')
        # Parse it
        pgn_id, source_id, dest, priority = NMEA2000Decoder._extract_header(frame_id_int)

        # Extract and reverse the CAN data
        can_data = packet[6:6 + data_length][::-1]
               
        # Log the extracted information including the combined string
        logger.debug("PGN ID: %s, Frame ID: %s, CAN Data: %s, Source ID: %s",
            pgn_id,
            binascii.hexlify(frame_id).decode('ascii'),
            can_data,
            source_id)
        
        return self._decode(pgn_id, priority, source_id, dest, datetime.now(), bytes(can_data))

    @staticmethod
    def _isFastPGN(pgn_id: int) -> bool:
        is_fast_func_name = f'is_fast_pgn_{pgn_id}'
        is_fast_func = globals().get(is_fast_func_name)

        if is_fast_func:
            is_fast = is_fast_func()
            logger.info(f"Is fast PGN: {is_fast}")
            return is_fast
        else:
            raise ValueError(f"No function found for PGN: {pgn_id}")

    def _decode(self, pgn: int, priority: int, source_id: int, destination_id: int, timestamp: datetime, can_data: bytes, already_combined: bool = False) -> NMEA2000Message | None:
        """Decode a single PGN message."""

        # Check if the PGN should be excluded or included
        if pgn in self.exclude_pgns:
            logger.debug(f"Excluding PGN: {pgn}")
            return None
        if len(self.include_pgns) > 0 and pgn not in self.include_pgns:
            logger.debug(f"Excluding (by include) PGN: {pgn}")
            return None

        source_iso_name = None
        if pgn != ISO_CLAIM_PGN: # The ISO_CLAIM_PGN should bypass this check so we can build the map later
            source_iso_name = self.source_to_iso_name.get(source_id, None)
            if source_iso_name is None and self.build_network_map:
                if self.started_at < datetime.now() - timedelta(minutes=10):
                    logger.warning("No ISO name found for source %s in PGN id %s. Skipping the message.", source_id, pgn)
                else:
                    logger.info("No ISO name found for source %s in PGN id %s. Skipping the message.", source_id, pgn)
                return None
        
            if source_iso_name is not None and source_iso_name.manufacturer_code is not None:
                # Check if the PGN should be excluded or included based on manufacturer
                manufacturer_code = source_iso_name.manufacturer_code.lower()
                if manufacturer_code in self.exclude_manufacturer_code:
                    logger.debug(f"Excluding PGN: {pgn} based on manufacturer code {source_iso_name.manufacturer_code}")
                    return None
                if len(self.include_manufacturer_code) > 0 and manufacturer_code not in self.include_manufacturer_code:
                    logger.debug(f"Excluding (by include) PGN: {pgn} based on manufacturer code {source_iso_name.manufacturer_code}")
                    return None

        is_fast = False
        if not already_combined:
            is_fast = NMEA2000Decoder._isFastPGN(pgn)

        if (is_fast):
            return self._decode_fast_message(pgn, priority, source_id, destination_id, timestamp, can_data, source_iso_name)
        else:
            return self._call_decode_function(pgn, priority, source_id, destination_id, timestamp, can_data, source_iso_name)

    def _call_decode_function(self, pgn:int, priority: int, src: int, dest: int, timestamp: datetime, data:bytes, source_iso_name: IsoName | None) -> NMEA2000Message | None:
        decode_func_name = f'decode_pgn_{pgn}'
        decode_func = globals().get(decode_func_name)

        if not decode_func:
            raise ValueError(f"No decoding function found for PGN: {pgn}")

        data_int = int.from_bytes(data, "big")
        nmea2000Message = decode_func(data_int)
        # Check if the PGN should be excluded or included by ID
        id = nmea2000Message.id.lower()
        if id in self.exclude_pgns_ids:
            logger.debug(f"Excluding PGN by id: {nmea2000Message.id}")
            return None
        if len(self.include_pgns_ids) > 0 and id not in self.include_pgns_ids:
            logger.debug(f"Excluding (by include) PGN by id: {nmea2000Message.id}")
            return None
        
        # Handle dump to file
        if (self.dump_TextIOWrapper is not None) and (len(self.dump_include_pgns)+len(self.dump_include_pgns_ids) == 0 or nmea2000Message.PGN in self.dump_include_pgns or nmea2000Message.id in self.dump_include_pgns_ids):
            str = nmea2000Message.to_json() + "\n"
            self.dump_TextIOWrapper.write(str)
        
        # Handle ISO Address Claim messages and enrichment
        if nmea2000Message.PGN == ISO_CLAIM_PGN:
            # In this message the data is a 64 bit unique NAME which is stable between network restarts
            source_iso_name = self.source_to_iso_name[src] = IsoName(nmea2000Message, data_int)
            if self.iso_claim_filter:
                return None
            
        nmea2000Message.add_data(src, dest, priority, timestamp, source_iso_name, self.build_network_map)
        nmea2000Message.apply_preferred_units(self.preferred_units)

        return nmea2000Message

    def close(self):
        """Close the dump_TextIOWrapper file if it is open."""
        if self.dump_TextIOWrapper:
            self.dump_TextIOWrapper.close()
            self.dump_TextIOWrapper = None
            logger.info("dump_TextIOWrapper file has been closed.")

    def __enter__(self):
        """Enter the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context and clean up resources."""
        self.close()



from .decoder import NMEA2000Decoder
from .encoder import NMEA2000Encoder
from .consts import FieldTypes, PhysicalQuantities
from .message import NMEA2000Message, NMEA2000Field, IsoName
from .ioclient import EByteNmea2000Gateway, ActisenseNmea2000Gateway, YachtDevicesNmea2000Gateway, WaveShareNmea2000Gateway, AsyncIOClient, State

__all__ = ["NMEA2000Decoder", "NMEA2000Encoder", "NMEA2000Message", "NMEA2000Field", "EByteNmea2000Gateway", "ActisenseNmea2000Gateway", "YachtDevicesNmea2000Gateway", "WaveShareNmea2000Gateway", "AsyncIOClient", "FieldTypes", "PhysicalQuantities", "State", "IsoName"]
